/*******************************************************************************
 *   Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2015 Peter Kolb
 *   peter.kolb@linguatools.org
 *
 *   Licensed under the Apache License, Version 2.0 (the "License"); you may not
 *   use this file except in compliance with the License. You may obtain a copy
 *   of the License at 
 *   
 *        http://www.apache.org/licenses/LICENSE-2.0 
 *
 *   Unless required by applicable law or agreed to in writing, software 
 *   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the 
 *   License for the specific language governing permissions and limitations
 *   under the License.
 *
 ******************************************************************************/

package de.linguatools.disco;

import java.io.IOException;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import org.apache.lucene.document.Document;
import org.apache.lucene.index.DirectoryReader;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.store.FSDirectory;

/**
 * This class contains methods that are based on the rank a word occupies in the
 * similarity list of other words.
 * @author peter
 * @version 2.0
 */
public class Rank {
   
    /***************************************************************************
     * Computes the rank of w2 in the similarity list of w1. If w2 does not
     * occur in the similarity list of w1, the return value is the highest rank 
     * + 1. (The highest rank is equal to the length of the list of similar words
     * stored for w1. The default value is 300. See 
     * <a href="http://www.linguatools.de/disco/disco-builder.html">DISCOBuilder</a>
     * option <i>-nBest</i>.)<br/>
     * If w1 is not in the index, 0 is returned.<br/>
     * <b>Important note</b>: This method does not work with word spaces of type
     * <code>DISCO.WordspaceType.COL</code> (use 
     * <code>de.linguatools.disco.Rank.rankCol</code> instead).
     * @param disco
     * @param w1 input word #1 (must be a single token).
     * @param w2 input word #2 (must be a single token).
     * @return rank (1...n) or 0
     * @throws IOException
     * @throws WrongWordspaceTypeException if the <code>disco</code> word space
     * is not of type <code>DISCO.WordspaceType.SIM</code>.
     */
    public int rankSim(DISCO disco, String w1, String w2) throws IOException, 
            WrongWordspaceTypeException {
        
        // check word space type
        if( disco.wordspaceType != DISCO.WordspaceType.SIM ){
            throw new WrongWordspaceTypeException("This method can not be applied"
                    + "to word spaces of type "+disco.wordspaceType);
        }
        
        Document doc = disco.searchIndex(w1);
        if (doc == null) {
            return 0;
        }
        ReturnDataBN res = new ReturnDataBN();
        res.words = doc.get("dsb").split(" ");
        for (int i = 0; i < res.words.length; i++) {
            if (res.words[i].equals(w2)) {
                return i + 1;
            }
        }
        return res.words.length + 1;
    }

    /***************************************************************************
     * Computes the rank of w2 among the collocations of w1. If w2 does not
     * occur as collocation of w1, the return value is the highest rank + 1. If
     * w1 is not found in the index, the return value is 0.<br/>
     * See <a href="DISCO.html#collocations(java.lang.String)">DISCO.collocations</a>
     * for more information. 
     * @param disco
     * @param w1 input word #1 (must be a single token).
     * @param w2 input word #2 (must be a single token).
     * @return rank (1...n) or 0
     * @throws IOException
     */
    public int rankCol(DISCO disco, String w1, String w2) throws IOException {
        
        ReturnDataCol[] res = disco.collocations(w1);
        if( res == null) return 0;
        for (int i = 0; i < res.length; i++) {
            if (res[i].word.equals(w2)) {
                return i + 1;
            }
        }
        // if w2 is not among the collocates of w1, return highest rank + 1
        return res.length + 1;
    }

    /***************************************************************************
     * Finds the words in the index in whose similarity or collocation lists the
     * <code>words</code> rank highest. There must be at least one word in
     * <code>words</code>.<br/>
     * For each word <code>v</code> from the vocabulary, the method looks up on
     * which rank <code>v</code> occurs among the similar words or collocations
     * of the words from the input set. The score of <code>v</code> is the
     * product of these ranks. The lower the rank product, the more similar
     * <code>v</code> is to <code>words</code>.<br/>
     * For low frequency words this method often gives better results than
     * de.linguatools.disco.similarWords().<br/>
     * <b>Warning</b>: This operation is very time consuming! It should only be
     * used if the word space has been loaded into main memory.
     * @param disco DISCO word space (in-memory)
     * @param words set of input words (all words must be single tokens).
     * @param type If type is <code>WordspaceType.SIM</code> then the list of the
     * semantically most similar words is searched, otherwise the list of 
     * significant co-occurrences (=collocations) is searched. Therefore, 
     * <code>WordspaceType.SIM</code> is only possible with word spaces of this
     * type!
     * @return ArrayList with similar words, most similar first. If none of the
     * <code>words</code> was found in the index, the return value is null.
     * @throws IOException
     * @throws WrongWordspaceTypeException
     */
    public ArrayList<WordAndRank> highestRanking(DISCO disco, Set<String> words, 
            DISCO.WordspaceType type)
            throws IOException, WrongWordspaceTypeException {

        // check word space type
        if( type == DISCO.WordspaceType.SIM ){
            if( disco.wordspaceType != DISCO.WordspaceType.SIM ){
                throw new WrongWordspaceTypeException("This method was called "
                        + "with parameter type = DISCO.WordspaceType.SIM, but the"
                        + "word space is of type "+disco.wordspaceType);
            }
        }
        
        ArrayList<WordAndRank> result = new ArrayList<>();

        // store words from the Set in Hash
        HashMap wordsHash = new HashMap();
        for(Iterator it = words.iterator(); it.hasNext(); ){
            wordsHash.put(it.next(), 1);
        }

        // open index reader and get number of words (Lucene documents) in the
        // index
        IndexReader reader;
        if (disco.indexRAM == null) {
            reader = DirectoryReader.open(FSDirectory.open(Paths.get(disco.indexDir)));
        } else {
            reader = DirectoryReader.open(disco.indexRAM);
        }
        int numDocs = reader.numDocs();

        // run trough list of all words in index
        for (int i = 0; i <= numDocs; i++) {
            Document doc;
            try{
                doc = reader.document(i);
            }catch(IOException ex){
                continue;
            }
            String word = doc.get("word");
            // leave out words from the input set
            if ( wordsHash.containsKey(word) ) {
                continue;
            }
            // look up rank of the words in the input Set in the list of the most
            // similar words (or collocations) of current term
            long rankProdukt = 1;
            long r;
            boolean w1Found = false;
            for (String w1 : words) {
                if( type == DISCO.WordspaceType.SIM){
                    r = rankSim(disco, w1, word);
                }else{
                    r = rankCol(disco, w1, word);
                }
                if( r != 0 ){
                    rankProdukt = rankProdukt * r;
                    w1Found = true;
                }
            }
            if( w1Found == false ){
                System.out.println("*** Error: none of the input words was " +
                        "found in the index.");
                return null;
            }else{
                result.add(new WordAndRank(word, rankProdukt));
            }
        }
        reader.close();
        // sort result Vector by rank (lowest first)
        Collections.sort(result);
        // return sorted Vector
        return result;
    }

    /***************************************************************************
     * data structure.
     */
    public class WordAndRank implements Comparable<WordAndRank>{

        String word;
        long rank;

        // Konstruktor
        public WordAndRank(String word, long rank) {
            this.word = word;
            this.rank = rank;
        }
        
        @Override
        public int compareTo(WordAndRank other) {
            int retval = 0;
            if (rank < other.rank) {
                retval = -1;
            }
            if (rank > other.rank) {
                retval = 1;
            }
            return (retval);
        }
        
    }

}
